#    Copyright 2020 Red Hat, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from ovsdbapp.backend.ovs_idl import event as row_event

# TODO(mjozefcz): Start consuming const and utils
# from neutron-lib once released.
from ovn_octavia_provider.common import constants as ovn_const

LOG = logging.getLogger(__name__)


class LogicalRouterPortEvent(row_event.RowEvent):

    def __init__(self, driver):
        table = 'Logical_Router_Port'
        events = (self.ROW_CREATE, self.ROW_DELETE)
        super().__init__(events, table, None)
        self.event_name = 'LogicalRouterPortEvent'
        self.driver = driver

    def run(self, event, row, old):
        LOG.debug('LogicalRouterPortEvent logged, '
                  '%(event)s, %(row)s',
                  {'event': event,
                   'row': row})
        if event == self.ROW_CREATE:
            self.driver.lb_create_lrp_assoc_handler(row)
        elif event == self.ROW_DELETE:
            self.driver.lb_delete_lrp_assoc_handler(row)


class LogicalSwitchPortUpdateEvent(row_event.RowEvent):

    def __init__(self, driver):
        table = 'Logical_Switch_Port'
        events = (self.ROW_UPDATE,)
        super().__init__(events, table, None)
        self.event_name = 'LogicalSwitchPortUpdateEvent'
        self.driver = driver

    def match_fn(self, event, row, old):
        port_name = row.external_ids.get(
            ovn_const.OVN_PORT_NAME_EXT_ID_KEY, '')
        if hasattr(old, 'external_ids') and port_name.startswith(
                ovn_const.LB_VIP_PORT_PREFIX):
            return True
        return False

    def run(self, event, row, old):
        LOG.debug('LogicalSwitchPortUpdateEvent logged, '
                  '%(event)s, %(row)s',
                  {'event': event,
                   'row': row})
        fip_old = old.external_ids.get(ovn_const.OVN_PORT_FIP_EXT_ID_KEY)
        fip_new = row.external_ids.get(ovn_const.OVN_PORT_FIP_EXT_ID_KEY)
        if fip_old != fip_new:
            if fip_old and fip_new is None:
                action = ovn_const.REQ_INFO_ACTION_DISASSOCIATE
                fip = fip_old
            else:
                action = ovn_const.REQ_INFO_ACTION_ASSOCIATE
                fip = fip_new
            self.driver.vip_port_update_handler(row, fip, action)


class ServiceMonitorUpdateEvent(row_event.RowEvent):

    def __init__(self, driver):
        table = 'Service_Monitor'
        events = (self.ROW_UPDATE, self.ROW_DELETE)
        super().__init__(events, table, None)
        self.event_name = 'ServiceMonitorUpdateEvent'
        self.driver = driver

    def run(self, event, row, old):
        LOG.debug('ServiceMonitorUpdateEvent logged, '
                  '%(event)s, %(row)s',
                  {'event': event,
                   'row': row})
        if event == self.ROW_DELETE:
            self.driver.sm_update_event_handler(row, sm_delete_event=True)
        elif event == self.ROW_UPDATE:
            self.driver.sm_update_event_handler(row)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import atexit
import multiprocessing as mp

from neutron.common import utils as n_utils
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import connection

from ovn_octavia_provider import agent as ovn_agent
from ovn_octavia_provider.common import config as ovn_config
from ovn_octavia_provider.common import constants as ovn_const
from ovn_octavia_provider import event as ovn_event
from ovn_octavia_provider import helper as ovn_helper
from ovn_octavia_provider.ovsdb import impl_idl_ovn
from ovn_octavia_provider.tests.functional import base as ovn_base


class TestOvnOctaviaProviderAgent(ovn_base.TestOvnOctaviaBase):

    def setUp(self):
        super().setUp()
        self._initialize_ovn_da()

    def _initialize_ovn_da(self):
        # NOTE(mjozefcz): In theory this is separate process
        # with IDL running, but to make it easier for now
        # we can initialize this IDL here instead spawning
        # another process.
        ovn_config.register_opts()
        da_helper = ovn_helper.OvnProviderHelper()
        events = [ovn_event.LogicalRouterPortEvent(da_helper),
                  ovn_event.LogicalSwitchPortUpdateEvent(da_helper)]
        ovn_nb_idl_for_events = impl_idl_ovn.OvnNbIdlForLb(
            event_lock_name='func_test')
        ovn_nb_idl_for_events.notify_handler.watch_events(events)
        c = connection.Connection(ovn_nb_idl_for_events,
                                  ovn_config.get_ovn_ovsdb_timeout())
        ovn_nbdb_api = impl_idl_ovn.OvsdbNbOvnIdl(c)
        atexit.register(ovn_nbdb_api.ovsdb_connection.stop)

    def _test_lrp_event_handler(self, cascade=False):
        # Create Network N1 on router R1 and LBA on N1
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)

        # Create Network N2, connect it to R1
        network_N2 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id2, subnet2_id = self._create_subnet_from_net(
            network_N2, '10.0.1.0/24')
        port_address2, port_id2 = self._create_port_on_network(network_N2)

        # Create Network N3
        network_N3 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id3, subnet3_id = self._create_subnet_from_net(
            network_N3, '10.0.2.0/24')
        port_address3, port_id3 = self._create_port_on_network(network_N3)

        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        self._attach_router_to_subnet(subnet2_id, r1_id)

        lba_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        # Check if LBA exists in N2 LS
        n_utils.wait_until_true(
            lambda: self._is_lb_associated_to_ls(
                lba_data['model'].loadbalancer_id,
                ovn_const.LR_REF_KEY_HEADER + network_id2),
            timeout=10)

        lbb_data = self._create_load_balancer_and_validate(
            network_id3, subnet3_id, port_address3, port_id3, multiple_lb=True)
        # Add N3 to R1
        self.l3_plugin.add_router_interface(
            self.context, lba_data[
                ovn_const.LB_EXT_IDS_LR_REF_KEY][
                    len(ovn_const.LR_REF_KEY_HEADER):],
            {'subnet_id': lbb_data['vip_net_info'][1]})

        # Check LBB exists on R1
        n_utils.wait_until_true(
            lambda: self._is_lb_associated_to_lr(
                lbb_data['model'].loadbalancer_id,
                lba_data[ovn_const.LB_EXT_IDS_LR_REF_KEY]),
            timeout=10)
        # Check LBA connected to N3
        n_utils.wait_until_true(
            lambda: self._is_lb_associated_to_ls(
                lba_data['model'].loadbalancer_id,
                ovn_const.LR_REF_KEY_HEADER + lbb_data['vip_net_info'][0]),
            timeout=10)
        # Check LBB connected to N1
        n_utils.wait_until_true(
            lambda: self._is_lb_associated_to_ls(
                lbb_data['model'].loadbalancer_id,
                ovn_const.LR_REF_KEY_HEADER + lba_data['vip_net_info'][0]),
            timeout=10)
        # Check LBB connected to N2
        n_utils.wait_until_true(
            lambda: self._is_lb_associated_to_ls(
                lbb_data['model'].loadbalancer_id,
                ovn_const.LR_REF_KEY_HEADER + network_id2),
            timeout=10)

        lbb_id = lbb_data['model'].loadbalancer_id
        if not cascade:
            # N3 removed from R1
            self.l3_plugin.remove_router_interface(
                self.context, lba_data[
                    ovn_const.LB_EXT_IDS_LR_REF_KEY][
                        len(ovn_const.LR_REF_KEY_HEADER):],
                {'subnet_id': lbb_data['vip_net_info'][1]})
        else:
            # Delete LBB Cascade
            self._delete_load_balancer_and_validate(lbb_data, cascade=True,
                                                    multiple_lb=True)

        # Check LBB doesn't exists on R1
        n_utils.wait_until_true(
            lambda: not self._is_lb_associated_to_lr(
                lbb_id, lba_data[ovn_const.LB_EXT_IDS_LR_REF_KEY]),
            timeout=10)
        # Check LBB not connected to N1
        n_utils.wait_until_true(
            lambda: not self._is_lb_associated_to_ls(
                lbb_id,
                ovn_const.LR_REF_KEY_HEADER + lba_data['vip_net_info'][0]),
            timeout=10)
        # Check LBB not connected to N2
        n_utils.wait_until_true(
            lambda: not self._is_lb_associated_to_ls(
                lbb_id, ovn_const.LR_REF_KEY_HEADER + network_id2),
            timeout=10)

    def test_lrp_event_handler_with_interface_delete(self):
        self._test_lrp_event_handler()

    def test_lrp_event_handler_with_loadbalancer_cascade_delete(self):
        self._test_lrp_event_handler(cascade=True)

    def test_lrp_event_handler_lrp_with_external_gateway(self):
        # Create Network N1 on router R1 and LBA on N1
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        lba_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        # Create provider network N2, connect it to R1
        provider_net, provider_subnet = self._create_provider_network()
        self.l3_plugin.update_router(
            self.context,
            r1_id,
            {'router': {
                'id': r1_id,
                'external_gateway_info': {
                    'enable_snat': True,
                    'network_id': provider_net['network']['id'],
                    'external_fixed_ips': [
                        {'ip_address': '100.0.0.2',
                         'subnet_id': provider_subnet['subnet']['id']}]}}})

        # Check if LBA doesn't exist in provider network LS
        n_utils.wait_until_true(
            lambda: not self._is_lb_associated_to_ls(
                lba_data['model'].loadbalancer_id,
                ovn_const.LR_REF_KEY_HEADER + provider_net['network']['id']),
            timeout=10)

    def test_fip_on_lb_vip(self):
        """This test checks if FIP on LB VIP is configured.

           This test validates if Load_Balancer VIP field
           consist Floating IP address that is configured
           on LB VIP port.
        """
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        # Create a pool
        self._create_pool_and_validate(lb_data, "p1")
        pool_id = lb_data['pools'][0].pool_id
        # Create listener
        self._create_listener_and_validate(lb_data, pool_id, 80)
        # Create Member-1 and associate it with lb_data
        self._create_member_and_validate(
            lb_data, pool_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')

        # Create provider network.
        e1, e1_s1 = self._create_provider_network()

        # Configure external_gateway for router
        router_id = lb_data['lr_ref'][8::]
        self.l3_plugin.update_router(
            self.context,
            router_id,
            {'router': {
                'id': router_id,
                'external_gateway_info': {
                    'enable_snat': True,
                    'network_id': e1['network']['id'],
                    'external_fixed_ips': [
                        {'ip_address': '100.0.0.2',
                         'subnet_id': e1_s1['subnet']['id']}]}}})

        # Create floating IP on LB VIP port
        vip_port_id = lb_data['model'].vip_port_id
        vip_port = self.core_plugin.get_ports(
            self.context, filters={'id': [vip_port_id]})[0]
        self.l3_plugin.create_floatingip(
            self.context, {'floatingip': {
                'tenant_id': self._tenant_id,
                'floating_network_id': e1['network']['id'],
                'subnet_id': None,
                'floating_ip_address': '100.0.0.20',
                'port_id': vip_port['id']}})

        # Validate if FIP is stored as VIP in LB
        lbs = self._get_loadbalancers()
        expected_vips = {
            '%s:80' % vip_port['fixed_ips'][0]['ip_address']: '10.0.0.10:80',
            '100.0.0.20:80': '10.0.0.10:80'}
        self.assertDictEqual(expected_vips,
                             lbs[0].get('vips'))

        provider_net = 'neutron-%s' % e1['network']['id']
        tenant_net = 'neutron-%s' % lb_data['model'].vip_network_id
        for ls in self.nb_api.tables['Logical_Switch'].rows.values():
            if ls.name == tenant_net:
                # Make sure that LB1 is added to tenant network
                self.assertIn(
                    lb_data['model'].loadbalancer_id,
                    [lb.name for lb in ls.load_balancer])
            elif ls.name == provider_net:
                # Make sure that LB1 is not added to provider net - e1 LS
                self.assertListEqual([], ls.load_balancer)

    def test_fip_on_lb_additional_vip(self):
        """This test checks if FIP on LB additional VIP is configured.

           This test validates if Load_Balancer additional VIP field
           consist Floating IP address that is configured
           on LB additional VIP port.
        """
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)

        network_ida, subnet_ida = self._create_subnet_from_net(
            network_N1, '10.0.1.0/24')
        port_addressa, port_ida = self._create_port_on_network(network_N1)
        additional_vips_list = [{
            'ip_address': port_addressa,
            'port_id': port_ida,
            'network_id': network_ida,
            'subnet_id': subnet_ida
        }]

        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        self._attach_router_to_subnet(subnet_ida, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id,
            additional_vips=additional_vips_list)
        # Create a pool
        self._create_pool_and_validate(lb_data, "p1")
        pool_id = lb_data['pools'][0].pool_id
        # Create listener
        self._create_listener_and_validate(lb_data, pool_id, 80)
        # Create Member-1 and associate it with lb_data
        self._create_member_and_validate(
            lb_data, pool_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')

        # Create provider network.
        e1, e1_s1 = self._create_provider_network()

        # Configure external_gateway for router
        self.l3_plugin.update_router(
            self.context,
            r1_id,
            {'router': {
                'id': r1_id,
                'external_gateway_info': {
                    'enable_snat': True,
                    'network_id': e1['network']['id'],
                    'external_fixed_ips': [
                        {'ip_address': '100.0.0.2',
                         'subnet_id': e1_s1['subnet']['id']}]}}})

        vip_port_id = lb_data['model'].vip_port_id
        vip_port = self.core_plugin.get_ports(
            self.context, filters={'id': [vip_port_id]})[0]

        # Create floating IP on LB additional VIP port
        addi_vip_port_id = lb_data['model'].additional_vips[0]['port_id']
        vipaddi_vip_port = self.core_plugin.get_ports(
            self.context, filters={'id': [addi_vip_port_id]})[0]
        self.l3_plugin.create_floatingip(
            self.context, {'floatingip': {
                'tenant_id': self._tenant_id,
                'floating_network_id': e1['network']['id'],
                'subnet_id': None,
                'floating_ip_address': '100.0.0.20',
                'port_id': vipaddi_vip_port['id']}})

        # Validate if FIP is stored as VIP in LB
        lbs = self._get_loadbalancers()
        expected_vips = {
            '%s:80' % vip_port['fixed_ips'][0]['ip_address']: '10.0.0.10:80',
            '%s:80' % vipaddi_vip_port['fixed_ips'][0]['ip_address']:
                '10.0.0.10:80',
            '100.0.0.20:80': '10.0.0.10:80'}
        self.assertDictEqual(expected_vips,
                             lbs[0].get('vips'))

        provider_net = 'neutron-%s' % e1['network']['id']
        tenant_net = 'neutron-%s' % lb_data['model'].vip_network_id
        for ls in self.nb_api.tables['Logical_Switch'].rows.values():
            if ls.name == tenant_net:
                # Make sure that LB1 is added to tenant network
                self.assertIn(
                    lb_data['model'].loadbalancer_id,
                    [lb.name for lb in ls.load_balancer])
            elif ls.name == provider_net:
                # Make sure that LB1 is not added to provider net - e1 LS
                self.assertListEqual([], ls.load_balancer)

    def test_agent_exit(self):
        exit_event = mp.Event()
        agent = mp.Process(target=ovn_agent.OvnProviderAgent,
                           args=[exit_event])
        agent.start()
        self.assertTrue(agent.is_alive())
        exit_event.set()
        agent.join()
        self.assertFalse(agent.is_alive())



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import os

from oslo_config import cfg
from oslo_log import log
from ovs.stream import Stream
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import event

from ovn_octavia_provider.common import config as ovn_config

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class BaseOvnIdl(connection.OvsdbIdl):
    @classmethod
    def from_server(cls, connection_string, schema_name):
        check_and_set_ssl_files(schema_name)
        helper = idlutils.get_schema_helper(connection_string, schema_name)
        helper.register_all()
        return cls(connection_string, helper)


class OvnIdl(BaseOvnIdl):

    def __init__(self, driver, remote, schema, notifier=True):
        super().__init__(remote, schema)
        self.driver = driver
        if notifier:
            self.notify_handler = OvnDbNotifyHandler(driver)
        else:
            self.notify_handler = None
        # ovsdb lock name to acquire.
        # This event lock is used to handle the notify events sent by idl.Idl
        # idl.Idl will call notify function for the "update" rpc method it
        # receives from the ovsdb-server.
        # This event lock is required for the following reasons
        #  - If there are multiple neutron servers running, OvnWorkers of
        #    these neutron servers would receive the notify events from
        #    idl.Idl
        #
        #  - we do not want all the neutron servers to handle these events
        #
        #  - only the neutron server which has the lock will handle the
        #    notify events.
        #
        #  - In case the neutron server which owns this lock goes down,
        #    ovsdb server would assign the lock to one of the other neutron
        #    servers.
        self.event_lock_name = "ovn_provider_driver_event_lock"

    def notify(self, event, row, updates=None):
        # Do not handle the notification if the event lock is requested,
        # but not granted by the ovsdb-server.
        if self.is_lock_contended:
            return
        if not self.notify_handler:
            return
        row = idlutils.frozen_row(row)
        self.notify_handler.notify(event, row, updates)

    @abc.abstractmethod
    def post_connect(self):
        """Should be called after the idl has been initialized"""


class OvnDbNotifyHandler(event.RowEventHandler):
    def __init__(self, driver):
        super().__init__()
        self.driver = driver


def check_and_set_ssl_files(schema_name):
    if schema_name in ['OVN_Northbound', 'OVN_Southbound']:
        if schema_name == 'OVN_Northbound':
            priv_key_file = ovn_config.get_ovn_nb_private_key()
            cert_file = ovn_config.get_ovn_nb_certificate()
            ca_cert_file = ovn_config.get_ovn_nb_ca_cert()
        else:
            priv_key_file = ovn_config.get_ovn_sb_private_key()
            cert_file = ovn_config.get_ovn_sb_certificate()
            ca_cert_file = ovn_config.get_ovn_sb_ca_cert()

        if priv_key_file:
            if not os.path.exists(priv_key_file):
                LOG.error(f"Cannot find private key file for {schema_name}")
            else:
                Stream.ssl_set_private_key_file(priv_key_file)

        if cert_file:
            if not os.path.exists(cert_file):
                LOG.error(f"Cannot find private cert file for {schema_name}")
            else:
                Stream.ssl_set_certificate_file(cert_file)

        if ca_cert_file:
            if not os.path.exists(ca_cert_file):
                LOG.error(f"Cannot find ca cert file for {schema_name}")
            else:
                Stream.ssl_set_ca_cert_file(ca_cert_file)
    else:
        LOG.error(f"{schema_name} not supported")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_log import log as logging
from ovn_octavia_provider.common import config as ovn_conf
from ovn_octavia_provider import driver

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def setup_conf():
    conf = cfg.CONF
    ovn_conf.register_opts()
    logging.register_options(CONF)

    try:
        CONF(project='octavia')
    except TypeError:
        LOG.error('Error parsing the configuration values. Please verify.')
        raise
    return conf


def main():
    """Main method for syncing Octavia LBs (OVN provider) with OVN NB DB.

    This script provides a utility for syncing the OVN Northbound Database
    with the Octavia database.

    """
    setup_conf()
    logging.setup(CONF, 'octavia_ovn_db_sync_util')

    # Method can be call like `octavia-ovn-db-sync-util --debug`
    LOG.info("OVN Octavia DB sync start.")
    args = sys.argv[1:]
    lb_filters = {'provider': 'ovn'}
    if '--debug' in args:
        cfg.CONF.set_override('debug', True)
        args.remove('--debug')
    else:
        cfg.CONF.set_override('debug', False)

    ovn_driver = driver.OvnProviderDriver()
    ovn_driver.do_sync(**lb_filters)
    LOG.info("OVN Octavia DB sync finish.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import atexit
import contextlib

import netaddr
from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc
from oslo_log import log
from ovsdbapp.backend import ovs_idl
from ovsdbapp.backend.ovs_idl import command
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.backend.ovs_idl import rowview
from ovsdbapp.backend.ovs_idl import transaction as idl_trans
from ovsdbapp.schema.ovn_northbound import impl_idl as nb_impl_idl
from ovsdbapp.schema.ovn_southbound import impl_idl as sb_impl_idl
import tenacity

from ovn_octavia_provider.common import config
from ovn_octavia_provider.common import exceptions as ovn_exc
from ovn_octavia_provider.common import utils
from ovn_octavia_provider.i18n import _
from ovn_octavia_provider.ovsdb import ovsdb_monitor


LOG = log.getLogger(__name__)


class OvnNbTransaction(idl_trans.Transaction):

    def __init__(self, *args, **kwargs):
        # NOTE(lucasagomes): The bump_nb_cfg parameter is only used by
        # the agents health status check
        self.bump_nb_cfg = kwargs.pop('bump_nb_cfg', False)
        super().__init__(*args, **kwargs)

    def pre_commit(self, txn):
        if not self.bump_nb_cfg:
            return
        self.api.nb_global.increment('nb_cfg')


class Backend(ovs_idl.Backend):

    def is_table_present(self, table_name):
        return table_name in self._tables

    def is_col_present(self, table_name, col_name):
        return self.is_table_present(table_name) and (
            col_name in self._tables[table_name].columns)

    # Check for a column match in the table. If not found do a retry with
    # a stop delay of 10 secs. This function would be useful if the caller
    # wants to verify for the presence of a particular row in the table
    # with the column match before doing any transaction.
    # Eg. We can check if Logical_Switch row is present before adding a
    # logical switch port to it.
    @tenacity.retry(retry=tenacity.retry_if_exception_type(RuntimeError),
                    wait=tenacity.wait_exponential(),
                    stop=tenacity.stop_after_delay(10),
                    reraise=True)
    def check_for_row_by_value_and_retry(self, table, column, match):
        try:
            idlutils.row_by_value(self.idl, table, column, match)
        except idlutils.RowNotFound as e:
            msg = (_("%(match)s does not exist in %(column)s of %(table)s")
                   % {'match': match, 'column': column, 'table': table})
            raise RuntimeError(msg) from e


class OvsdbConnectionUnavailable(n_exc.ServiceUnavailable):
    message = _("OVS database connection to %(db_schema)s failed with error: "
                "'%(error)s'. Verify that the OVS and OVN services are "
                "available and that the 'ovn_nb_connection' and "
                "'ovn_sb_connection' configuration options are correct.")


class FindLbInTableCommand(command.ReadOnlyCommand):
    def __init__(self, api, lb, table):
        super().__init__(api)
        self.lb = lb
        self.table = table

    def run_idl(self, txn):
        self.result = [
            rowview.RowView(item) for item in
            self.api.tables[self.table].rows.values()
            if self.lb in item.load_balancer]


class GetLrsCommand(command.ReadOnlyCommand):
    def run_idl(self, txn):
        self.result = [
            rowview.RowView(item) for item in
            self.api.tables['Logical_Router'].rows.values()]


# NOTE(froyo): remove this class once ovsdbapp manages the IPv6 into [ ]
# https://bugs.launchpad.net/ovsdbapp/+bug/2057471
class DelBackendFromIPPortMapping(command.BaseCommand):
    table = 'Load_Balancer'

    def __init__(self, api, lb, backend_ip):
        super().__init__(api)
        self.lb = lb
        if netaddr.IPNetwork(backend_ip).version == n_const.IP_VERSION_6:
            self.backend_ip = f'[{backend_ip}]'
        else:
            self.backend_ip = backend_ip

    def run_idl(self, txn):
        try:
            ovn_lb = self.api.lookup(self.table, self.lb)
            ovn_lb.delkey('ip_port_mappings', self.backend_ip)
        except Exception:
            LOG.exception("Error deleting backend %s from ip_port_mappings "
                          "for LB uuid %s", str(self.backend_ip), str(self.lb))


# NOTE(froyo): remove this class once ovsdbapp manages the IPv6 into [ ]
# https://bugs.launchpad.net/ovsdbapp/+bug/2057471
class AddBackendToIPPortMapping(command.BaseCommand):
    table = 'Load_Balancer'

    def __init__(self, api, lb, backend_ip, port_name, src_ip):
        super().__init__(api)
        self.lb = lb
        self.backend_ip = backend_ip
        self.port_name = port_name
        self.src_ip = src_ip
        if netaddr.IPNetwork(backend_ip).version == n_const.IP_VERSION_6:
            self.backend_ip = f'[{backend_ip}]'
            self.src_ip = f'[{src_ip}]'

    def run_idl(self, txn):
        try:
            lb = self.api.lookup(self.table, self.lb)
            lb.setkey('ip_port_mappings', self.backend_ip,
                      '%s:%s' % (self.port_name, self.src_ip))
        except Exception:
            LOG.exception("Error adding backend %s to ip_port_mappings "
                          "for LB uuid %s", str(self.backend_ip), str(self.lb))


class OvsdbNbOvnIdl(nb_impl_idl.OvnNbApiIdlImpl, Backend):
    def __init__(self, connection):
        super().__init__(connection)
        self.idl._session.reconnect.set_probe_interval(
            config.get_ovn_ovsdb_probe_interval())

    @property
    def nb_global(self):
        return next(iter(self.tables['NB_Global'].rows.values()))

    def create_transaction(self, check_error=False, log_errors=True,
                           bump_nb_cfg=False):
        return OvnNbTransaction(
            self, self.ovsdb_connection, self.ovsdb_connection.timeout,
            check_error, log_errors, bump_nb_cfg=bump_nb_cfg)

    @contextlib.contextmanager
    def transaction(self, *args, **kwargs):
        """A wrapper on the ovsdbapp transaction to work with revisions.

        This method is just a wrapper around the ovsdbapp transaction
        to handle revision conflicts correctly.
        """
        try:
            with super().transaction(*args, **kwargs) as t:
                yield t
        except ovn_exc.RevisionConflict as e:
            LOG.info('Transaction aborted. Reason: %s', e)

    def find_lb_in_table(self, lb, table):
        return FindLbInTableCommand(self, lb, table)

    def get_lrs(self):
        return GetLrsCommand(self)

    # NOTE(froyo): remove this method once ovsdbapp manages the IPv6 into [ ]
    def lb_del_ip_port_mapping(self, lb_uuid, backend_ip):
        return DelBackendFromIPPortMapping(self, lb_uuid, backend_ip)

    # NOTE(froyo): remove this method once ovsdbapp manages the IPv6 into [ ]
    def lb_add_ip_port_mapping(self, lb_uuid, backend_ip, port_name, src_ip):
        return AddBackendToIPPortMapping(self, lb_uuid, backend_ip, port_name,
                                         src_ip)


class OvsdbSbOvnIdl(sb_impl_idl.OvnSbApiIdlImpl, Backend):
    def __init__(self, connection):
        super().__init__(connection)
        self.idl._session.reconnect.set_probe_interval(
            config.get_ovn_ovsdb_probe_interval())


class OvnNbIdlForLb(ovsdb_monitor.OvnIdl):
    SCHEMA = "OVN_Northbound"
    TABLES = ('Logical_Switch', 'Load_Balancer', 'Load_Balancer_Health_Check',
              'Logical_Router', 'Logical_Switch_Port', 'Logical_Router_Port',
              'Gateway_Chassis', 'NAT')

    def __init__(self, event_lock_name=None, notifier=True):
        self.conn_string = config.get_ovn_nb_connection()
        ovsdb_monitor.check_and_set_ssl_files(self.SCHEMA)
        helper = self._get_ovsdb_helper(self.conn_string)
        for table in OvnNbIdlForLb.TABLES:
            helper.register_table(table)
        super().__init__(
            driver=None, remote=self.conn_string, schema=helper,
            notifier=notifier)
        self.event_lock_name = event_lock_name
        if self.event_lock_name:
            self.set_lock(self.event_lock_name)

    @utils.retry()
    def _get_ovsdb_helper(self, connection_string):
        return idlutils.get_schema_helper(connection_string, self.SCHEMA)


class OvnSbIdlForLb(ovsdb_monitor.OvnIdl):
    SCHEMA = "OVN_Southbound"
    TABLES = ('Load_Balancer', 'Service_Monitor')

    def __init__(self, event_lock_name=None):
        self.conn_string = config.get_ovn_sb_connection()
        ovsdb_monitor.check_and_set_ssl_files(self.SCHEMA)
        helper = self._get_ovsdb_helper(self.conn_string)
        for table in OvnSbIdlForLb.TABLES:
            helper.register_table(table)
        super().__init__(
            driver=None, remote=self.conn_string, schema=helper)
        self.event_lock_name = event_lock_name
        if self.event_lock_name:
            self.set_lock(self.event_lock_name)
        atexit.register(self.stop)

    @utils.retry()
    def _get_ovsdb_helper(self, connection_string):
        return idlutils.get_schema_helper(connection_string, self.SCHEMA)

    def start(self):
        self.conn = connection.Connection(
            self, timeout=config.get_ovn_ovsdb_timeout())
        return OvsdbSbOvnIdl(self.conn)

    def stop(self):
        # Close the running connection if it has been initalized
        if hasattr(self, 'conn'):
            if not self.conn.stop(timeout=config.get_ovn_ovsdb_timeout()):
                LOG.debug("Connection terminated to OvnSb "
                          "but a thread is still alive")
            del self.conn
        # complete the shutdown for the event handler
        self.notify_handler.shutdown()
        # Close the idl session
        self.close()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest import mock

from keystoneauth1 import exceptions as ks_exceptions
from octavia_lib.api.drivers import exceptions as driver_exceptions
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslotest import base

from ovn_octavia_provider.common import clients
from ovn_octavia_provider.common import config


class TestKeystoneSession(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        config.register_opts()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))

    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options')
    def test_auth(self, kl_auth):
        missing_options = [mock.Mock(dest='username')]
        auth = mock.Mock()

        # service_auth with missing option
        kl_auth.side_effect = [
            ks_exceptions.auth_plugins.MissingRequiredOptions(missing_options)
        ]

        ksession = clients.KeystoneSession()
        self.assertRaises(
            ks_exceptions.auth_plugins.MissingRequiredOptions,
            lambda: ksession.auth)

        # neutron with missing option, missing option also in service_auth
        kl_auth.reset_mock()
        kl_auth.side_effect = [
            ks_exceptions.auth_plugins.MissingRequiredOptions(missing_options),
            auth,
            ks_exceptions.auth_plugins.MissingRequiredOptions(missing_options),
        ]

        ksession = clients.KeystoneSession('neutron')
        self.assertRaises(
            ks_exceptions.auth_plugins.MissingRequiredOptions,
            lambda: ksession.auth)

        # neutron with missing option, it is copied from service_auth
        kl_auth.reset_mock()
        kl_auth.side_effect = [
            ks_exceptions.auth_plugins.MissingRequiredOptions(missing_options),
            auth,
            auth,
        ]

        self.conf.config(group='service_auth',
                         endpoint_override='foo')

        ksession = clients.KeystoneSession('neutron')
        self.assertEqual(ksession.auth, auth)
        self.assertEqual(cfg.CONF.neutron.endpoint_override, 'foo')

    @mock.patch(
        'keystoneauth1.loading.load_session_from_conf_options')
    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options')
    def test_cached_session(self, kl_auth, kl_session):
        ksession = clients.KeystoneSession('neutron')
        self.assertIs(
            ksession.session,
            ksession.session)
        kl_session.assert_called_once_with(
            mock.ANY, 'neutron', auth=ksession.auth)

    @mock.patch(
        'keystoneauth1.loading.load_auth_from_conf_options')
    def test_cached_auth(self, kl):
        ksession = clients.KeystoneSession('neutron')
        self.assertIs(
            ksession.auth,
            ksession.auth)
        kl.assert_called_once_with(mock.ANY, 'neutron')


class TestNeutronAuth(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        config.register_opts()
        self.mock_client = mock.patch(
            'openstack.connection.Connection').start()
        clients.Singleton._instances = {}

    @mock.patch.object(clients, 'KeystoneSession')
    def test_init(self, mock_ks):
        clients.NeutronAuth()
        self.mock_client.assert_called_once_with(
            session=mock_ks().session)

    def test_singleton(self):
        c1 = clients.NeutronAuth()
        c2 = clients.NeutronAuth()
        self.assertIs(c1, c2)

    def test_singleton_exception(self):
        mock_client = mock.Mock()
        mock_client.network_proxy = 'foo'
        with mock.patch(
            'openstack.connection.Connection',
                side_effect=[RuntimeError,
                             mock_client, mock_client]) as os_sdk:
            self.assertRaises(
                RuntimeError,
                clients.NeutronAuth)
            c2 = clients.NeutronAuth()
            c3 = clients.NeutronAuth()
            self.assertIs(c2, c3)
            self.assertEqual(os_sdk._mock_call_count, 2)

    @mock.patch.object(clients, 'KeystoneSession')
    def test_get_client(self, mock_ks):
        clients.get_neutron_client()
        self.mock_client.assert_called_once_with(
            session=mock_ks().session)

    @mock.patch.object(clients, 'NeutronAuth', side_effect=[RuntimeError])
    def test_get_client_error(self, mock_ks):
        msg = self.assertRaises(
            driver_exceptions.DriverError,
            clients.get_neutron_client)
        self.assertEqual("An unknown driver error occurred.", str(msg))


class TestOctaviaAuth(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        config.register_opts()
        self.mock_client = mock.patch(
            'openstack.connection.Connection').start()
        clients.Singleton._instances = {}

    @mock.patch.object(clients, 'KeystoneSession')
    @mock.patch('openstack.connection.Connection')
    def test_init(self, mock_conn, mock_ks):
        clients.OctaviaAuth()
        mock_conn.assert_called_once_with(
            session=mock_ks().session,
            region_name=mock.ANY
        )

    def test_singleton(self):
        c1 = clients.OctaviaAuth()
        c2 = clients.OctaviaAuth()
        self.assertIs(c1, c2)

    def test_singleton_exception(self):
        mock_client = mock.Mock()
        mock_client.lbaas_proxy = 'foo'
        with mock.patch(
            'openstack.connection.Connection',
                side_effect=[RuntimeError,
                             mock_client, mock_client]) as os_sdk:
            self.assertRaises(
                RuntimeError,
                clients.OctaviaAuth)
            c2 = clients.OctaviaAuth()
            c3 = clients.OctaviaAuth()
            self.assertIs(c2, c3)
            self.assertEqual(os_sdk._mock_call_count, 2)

    @mock.patch.object(clients, 'KeystoneSession')
    @mock.patch('openstack.connection.Connection')
    def test_get_client(self, mock_conn, mock_ks):
        clients.get_octavia_client()
        mock_conn.assert_called_once_with(
            session=mock_ks().session,
            region_name=mock.ANY
        )

    @mock.patch.object(clients, 'OctaviaAuth', side_effect=[RuntimeError])
    def test_get_client_error(self, mock_ks):
        msg = self.assertRaises(
            driver_exceptions.DriverError,
            clients.get_octavia_client)
        self.assertEqual("An unknown driver error occurred.", str(msg))



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const
from octavia_lib.api.drivers import exceptions as o_exceptions
from octavia_lib.common import constants as o_constants
from oslo_utils import uuidutils

from ovn_octavia_provider.tests.functional import base as ovn_base


class TestOvnOctaviaProviderDriver(ovn_base.TestOvnOctaviaBase):

    def test_loadbalancer(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)

        network_ida, subnet_ida = self._create_subnet_from_net(
            network_N1, '2001:db8:0:1::/64', ip_version=n_const.IP_VERSION_6)
        port_addressa, port_ida = self._create_port_on_network(network_N1)
        additional_vips_list = [{
            'ip_address': port_addressa,
            'port_id': port_ida,
            'network_id': network_ida,
            'subnet_id': subnet_ida
        }]
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        self._attach_router_to_subnet(subnet_ida, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id,
            additional_vips=additional_vips_list)
        self._update_load_balancer_and_validate(lb_data, admin_state_up=False)
        self._update_load_balancer_and_validate(lb_data, admin_state_up=True)
        self._delete_load_balancer_and_validate(lb_data)
        # create load balance with admin state down
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id,
            admin_state_up=False)
        self._delete_load_balancer_and_validate(lb_data)

    def test_create_lb_custom_network(self):
        self._create_load_balancer_custom_lr_ls_and_validate(
            admin_state_up=True, create_router=True,
            force_retry_ls_to_lr_assoc=False)

    def test_create_lb_custom_network_retry(self):
        self._create_load_balancer_custom_lr_ls_and_validate(
            admin_state_up=True, create_router=True,
            force_retry_ls_to_lr_assoc=True)

    def test_delete_lb_on_nonexisting_lb(self):
        # LoadBalancer doesnt exist anymore, so just create a model and delete
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '19.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id,
            only_model=True)
        self.ovn_driver.loadbalancer_delete(lb_data['model'])
        expected_status = {
            'loadbalancers': [{"id": lb_data['model'].loadbalancer_id,
                               "provisioning_status": "DELETED",
                               "operating_status": "OFFLINE"}],
            'listeners': [],
            'pools': [],
            'members': [],
        }
        del lb_data['model']
        self._wait_for_status_and_validate(lb_data, [expected_status])

    def test_pool(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)
        self._create_pool_and_validate(lb_data, "p_TCP_1", protocol='TCP')
        self._update_pool_and_validate(lb_data, "p_TCP_1")
        self._create_pool_and_validate(lb_data, "p_UDP_1", protocol='UDP')
        self._create_pool_and_validate(lb_data, "p_SCTP_1", protocol='SCTP')
        self._create_pool_and_validate(lb_data, "p_TCP_2", protocol='TCP')
        self._update_pool_and_validate(lb_data, "p_TCP_2",
                                       admin_state_up=False)
        self._update_pool_and_validate(lb_data, "p_TCP_2",
                                       admin_state_up=True)
        self._update_pool_and_validate(lb_data, "p_TCP_2",
                                       admin_state_up=False)
        self._create_pool_and_validate(lb_data, "p_UDP_2", protocol='UDP')
        self._create_pool_and_validate(lb_data, "p_SCTP_2", protocol='SCTP')
        self._delete_pool_and_validate(lb_data, "p_SCTP_1")
        self._delete_pool_and_validate(lb_data, "p_UDP_1")
        self._delete_pool_and_validate(lb_data, "p_TCP_1")
        self._delete_load_balancer_and_validate(lb_data)

    def test_member(self):
        net1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            net1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(net1)
        net2 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id2, subnet_id2 = self._create_subnet_from_net(
            net2, '20.0.0.0/24')
        self._create_port_on_network(net2)

        net3 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id3, subnet_id3 = self._create_subnet_from_net(
            net3, '30.0.0.0/24')
        self._create_port_on_network(net3)

        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        self._attach_router_to_subnet(subnet_id2, r1_id)
        self._attach_router_to_subnet(subnet_id3, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        # TCP Pool
        self._create_pool_and_validate(lb_data, "p_TCP", protocol='TCP')

        # UDP Pool
        self._create_pool_and_validate(lb_data, "p_UDP", protocol='UDP')

        # SCTP Pool
        self._create_pool_and_validate(lb_data, "p_SCTP", protocol='SCTP')

        pool_TCP_id = lb_data['pools'][0].pool_id
        pool_UDP_id = lb_data['pools'][1].pool_id
        pool_SCTP_id = lb_data['pools'][2].pool_id

        # Members for TCP Pool
        self._create_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')
        self._update_member_and_validate(lb_data, pool_TCP_id, "10.0.0.10")
        self._update_member_and_validate(lb_data, pool_TCP_id, "10.0.0.10",
                                         admin_state_up=False)
        self._create_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.11')

        # Members for UDP Pool
        self._create_member_and_validate(
            lb_data, pool_UDP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')
        self._update_member_and_validate(lb_data, pool_UDP_id, "10.0.0.10")
        self._update_member_and_validate(lb_data, pool_UDP_id, "10.0.0.10",
                                         admin_state_up=False)
        self._create_member_and_validate(
            lb_data, pool_UDP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.11')

        # Members for SCTP Pool
        self._create_member_and_validate(
            lb_data, pool_SCTP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')
        self._update_member_and_validate(lb_data, pool_SCTP_id, "10.0.0.10")
        self._update_member_and_validate(lb_data, pool_SCTP_id, "10.0.0.10",
                                         admin_state_up=False)
        self._create_member_and_validate(
            lb_data, pool_SCTP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.11')

        # Disable loadbalancer
        self._update_load_balancer_and_validate(lb_data,
                                                admin_state_up=False)
        # Enable loadbalancer back
        self._update_load_balancer_and_validate(lb_data,
                                                admin_state_up=True)

        # Delete members from TCP Pool
        self._delete_member_and_validate(lb_data, pool_TCP_id,
                                         lb_data['vip_net_info'][0],
                                         '10.0.0.10')
        self._delete_member_and_validate(lb_data, pool_TCP_id,
                                         lb_data['vip_net_info'][0],
                                         '10.0.0.11')
        # Add again member to TCP Pool
        self._create_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')

        self._create_member_and_validate(lb_data, pool_TCP_id, subnet_id2,
                                         network_id2, '20.0.0.4')
        self._create_member_and_validate(lb_data, pool_TCP_id, subnet_id2,
                                         network_id2, '20.0.0.6')

        self._create_member_and_validate(lb_data, pool_TCP_id, subnet_id3,
                                         network_id3, '30.0.0.6')
        self._delete_member_and_validate(lb_data, pool_TCP_id, network_id2,
                                         '20.0.0.6')

        # Deleting the pool should also delete the members.
        self._delete_pool_and_validate(lb_data, "p_TCP")

        # Delete the whole LB.
        self._delete_load_balancer_and_validate(lb_data)

    def test_member_no_subnet(self):
        self._o_driver_lib.get_pool.return_value = None

        # Test creating Member without subnet and unknown pool
        m_member = self._create_member_model('pool_from_nowhere',
                                             None,
                                             '30.0.0.7', 80)
        self.assertRaises(o_exceptions.UnsupportedOptionError,
                          self.ovn_driver.member_create, m_member)

        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        # TCP Pool
        self._create_pool_and_validate(lb_data, "p_TCP", protocol='TCP')
        pool_TCP_id = lb_data['pools'][0].pool_id

        self._o_driver_lib.get_pool.return_value = lb_data['pools'][0]
        self._o_driver_lib.get_loadbalancer.return_value = lb_data['model']

        # Test deleting a member without subnet
        self._create_member_and_validate(
            lb_data, pool_TCP_id, None,
            lb_data['vip_net_info'][0], '10.0.0.10',
            expected_subnet=lb_data['vip_net_info'][1])
        self._delete_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][0],
            '10.0.0.10', remove_subnet_id=True)

        # Test update member without subnet
        self._create_member_and_validate(
            lb_data, pool_TCP_id, None,
            lb_data['vip_net_info'][0], '10.0.0.10',
            expected_subnet=lb_data['vip_net_info'][1])
        self._update_member_and_validate(
            lb_data, pool_TCP_id, "10.0.0.10", remove_subnet_id=True)

        # Test creating a Member without subnet but with pool
        self._create_member_and_validate(
            lb_data, pool_TCP_id, None,
            lb_data['vip_net_info'][0], '10.0.0.11',
            expected_subnet=lb_data['vip_net_info'][1])

        # Deleting the pool should also delete the members.
        self._delete_pool_and_validate(lb_data, "p_TCP")

        # Delete the whole LB.
        self._delete_load_balancer_and_validate(lb_data)

    def test_hm(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)

        network_ida, subnet_ida = self._create_subnet_from_net(
            network_N1, '2001:db8:0:1::/64', ip_version=n_const.IP_VERSION_6)
        port_addressa, port_ida = self._create_port_on_network(network_N1)
        additional_vips_list = [{
            'ip_address': port_addressa,
            'port_id': port_ida,
            'network_id': network_ida,
            'subnet_id': subnet_ida
        }]

        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        self._attach_router_to_subnet(subnet_ida, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id,
            additional_vips=additional_vips_list)

        # TCP Pool
        self._create_pool_and_validate(lb_data, "p_TCP",
                                       protocol=o_constants.PROTOCOL_TCP)

        pool_TCP_id = lb_data['pools'][0].pool_id

        # Members for TCP Pool
        self._create_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')
        self._create_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.11')

        # CRUD health monitor.
        self._create_hm_and_validate(lb_data, pool_TCP_id, 'hm_TCP', 10, 30, 3,
                                     o_constants.HEALTH_MONITOR_TCP)
        self._update_hm_and_validate(lb_data, pool_TCP_id,
                                     admin_state_up=False)
        self._update_hm_and_validate(lb_data, pool_TCP_id, admin_state_up=True)
        self._delete_hm_and_validate(lb_data, pool_TCP_id)

        # Create to test delete over pool
        self._create_hm_and_validate(lb_data, pool_TCP_id, 'hm_TCP', 10, 30, 3,
                                     o_constants.HEALTH_MONITOR_TCP)
        # Deleting the pool should also delete the health monitor.
        self._delete_pool_and_validate(lb_data, "p_TCP")

    def test_listener(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.1.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        network_N2 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id2, subnet_id2 = self._create_subnet_from_net(
            network_N2, '20.1.0.0/24')
        self._create_port_on_network(network_N2)

        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)
        self._attach_router_to_subnet(subnet_id2, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)
        self._create_pool_and_validate(lb_data, "p_TCP", protocol='TCP')
        self._create_pool_and_validate(lb_data, "p_UDP", protocol='UDP')
        self._create_pool_and_validate(lb_data, "p_SCTP", protocol='SCTP')
        pool_TCP_id = lb_data['pools'][0].pool_id
        pool_UDP_id = lb_data['pools'][1].pool_id
        pool_SCTP_id = lb_data['pools'][2].pool_id

        # Create member in TCP pool
        self._create_member_and_validate(
            lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.1.0.4')
        self._create_member_and_validate(
            lb_data, pool_TCP_id, subnet_id2, network_id2, '20.1.0.4',)

        # Create member in UDP pool
        self._create_member_and_validate(
            lb_data, pool_UDP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.1.0.5')
        self._create_member_and_validate(lb_data, pool_UDP_id,
                                         subnet_id2, network_id2,
                                         '20.1.0.5')

        # Create member in SCTP pool
        self._create_member_and_validate(
            lb_data, pool_SCTP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.1.0.6')
        self._create_member_and_validate(lb_data, pool_SCTP_id,
                                         subnet_id2, network_id2,
                                         '20.1.0.6')

        # Play around first listener linked to TCP pool
        self._create_listener_and_validate(
            lb_data, pool_TCP_id, 80, protocol='TCP')
        self._update_listener_and_validate(lb_data, protocol_port=80)
        self._update_listener_and_validate(
            lb_data, protocol_port=80, admin_state_up=True)
        self._update_listener_and_validate(
            lb_data, protocol_port=80, admin_state_up=False)
        self._update_listener_and_validate(
            lb_data, protocol_port=80, admin_state_up=True)
        self._create_listener_and_validate(
            lb_data, pool_TCP_id, protocol_port=82, protocol='TCP')

        # Play around second listener linked to UDP pool
        self._create_listener_and_validate(
            lb_data, pool_UDP_id, 53, protocol='UDP')
        self._update_listener_and_validate(lb_data, 53, protocol='UDP')
        self._update_listener_and_validate(
            lb_data, protocol_port=53, protocol='UDP', admin_state_up=True)
        self._update_listener_and_validate(
            lb_data, protocol_port=53, protocol='UDP', admin_state_up=False)
        self._update_listener_and_validate(
            lb_data, protocol_port=53, protocol='UDP', admin_state_up=True)
        self._create_listener_and_validate(
            lb_data, pool_UDP_id, protocol_port=21, protocol='UDP')

        # Play around third listener linked to SCTP pool
        self._create_listener_and_validate(
            lb_data, pool_SCTP_id, 8081, protocol='SCTP')
        self._update_listener_and_validate(lb_data, 8081, protocol='SCTP')
        self._update_listener_and_validate(
            lb_data, protocol_port=8081, protocol='SCTP', admin_state_up=True)
        self._update_listener_and_validate(
            lb_data, protocol_port=8081, protocol='SCTP', admin_state_up=False)
        self._update_listener_and_validate(
            lb_data, protocol_port=8081, protocol='SCTP', admin_state_up=True)
        self._create_listener_and_validate(
            lb_data, pool_SCTP_id, protocol_port=8082, protocol='SCTP')

        # Delete listeners linked to TCP pool
        self._delete_listener_and_validate(
            lb_data, protocol_port=82, protocol='TCP')
        self._delete_listener_and_validate(
            lb_data, protocol_port=80, protocol='TCP')
        # Delete TCP pool members
        self._delete_member_and_validate(lb_data, pool_TCP_id,
                                         network_id2, '20.1.0.4')
        self._delete_member_and_validate(lb_data, pool_TCP_id,
                                         lb_data['vip_net_info'][0],
                                         '10.1.0.4')
        # Delete empty, TCP pool
        self._delete_pool_and_validate(lb_data, "p_TCP")
        # Delete the rest
        self._delete_load_balancer_and_validate(lb_data)

    def _test_cascade_delete(self, pool=True, listener=True, member=True):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        if pool:
            self._create_pool_and_validate(lb_data, "p_TCP", protocol='TCP')
            self._create_pool_and_validate(lb_data, "p_UDP", protocol='UDP')
            self._create_pool_and_validate(lb_data, "p_SCTP", protocol='SCTP')
            pool_TCP_id = lb_data['pools'][0].pool_id
            pool_UDP_id = lb_data['pools'][1].pool_id
            pool_SCTP_id = lb_data['pools'][2].pool_id
            if member:
                self._create_member_and_validate(
                    lb_data, pool_TCP_id, lb_data['vip_net_info'][1],
                    lb_data['vip_net_info'][0], '10.0.0.10')
                self._create_member_and_validate(
                    lb_data, pool_UDP_id, lb_data['vip_net_info'][1],
                    lb_data['vip_net_info'][0], '10.0.0.10')
                self._create_member_and_validate(
                    lb_data, pool_SCTP_id, lb_data['vip_net_info'][1],
                    lb_data['vip_net_info'][0], '10.0.0.10')
            if listener:
                self._create_listener_and_validate(
                    lb_data, pool_TCP_id, protocol_port=80, protocol='TCP')
                self._create_listener_and_validate(
                    lb_data, pool_UDP_id, protocol_port=53, protocol='UDP')
                self._create_listener_and_validate(
                    lb_data, pool_SCTP_id, protocol_port=8081, protocol='SCTP')

        self._delete_load_balancer_and_validate(lb_data, cascade=True)

    def test_lb_listener_pools_cascade(self):
        self._test_cascade_delete(member=False)

    def test_lb_pool_cascade(self):
        self._test_cascade_delete(member=False, listener=False)

    def test_cascade_delete(self):
        self._test_cascade_delete()

    def test_hm_unsupported_protocol(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        self._create_pool_and_validate(lb_data, "p_SCTP",
                                       protocol=o_constants.PROTOCOL_SCTP)
        pool_SCTP_id = lb_data['pools'][0].pool_id
        self._create_member_and_validate(
            lb_data, pool_SCTP_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')
        self.assertRaises(o_exceptions.UnsupportedOptionError,
                          self._create_hm_and_validate, lb_data, pool_SCTP_id,
                          'hm_SCTP', 10, 30, 3,
                          o_constants.HEALTH_MONITOR_SCTP)
        self._delete_load_balancer_and_validate(lb_data)

    def test_for_unsupported_options(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)

        m_pool = self._create_pool_model(lb_data['model'].loadbalancer_id,
                                         'lb1')
        m_pool.protocol = o_constants.PROTOCOL_HTTP
        self.assertRaises(o_exceptions.UnsupportedOptionError,
                          self.ovn_driver.pool_create, m_pool)

        self.assertRaises(o_exceptions.UnsupportedOptionError,
                          self.ovn_driver.loadbalancer_failover,
                          lb_data['model'].loadbalancer_id)

        m_listener = self._create_listener_model(
            lb_data['model'].loadbalancer_id, m_pool.pool_id, 80)
        m_listener.protocol = o_constants.PROTOCOL_HTTP
        self.assertRaises(o_exceptions.UnsupportedOptionError,
                          self.ovn_driver.listener_create, m_listener)
        self._create_listener_and_validate(lb_data)
        self._delete_load_balancer_and_validate(lb_data)

    def test_lb_listener_pool_workflow(self):
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)
        self._create_listener_and_validate(lb_data)
        self._create_pool_and_validate(
            lb_data, "p1", listener_id=lb_data['listeners'][0].listener_id)
        self._delete_pool_and_validate(
            lb_data, "p1", listener_id=lb_data['listeners'][0].listener_id)
        self._delete_listener_and_validate(lb_data)
        self._delete_load_balancer_and_validate(lb_data)

    def test_lb_member_batch_update(self):
        # Create a LoadBalancer
        network_N1 = self._create_net('N' + uuidutils.generate_uuid()[:4])
        network_id, subnet_id = self._create_subnet_from_net(
            network_N1, '10.0.0.0/24')
        port_address, port_id = self._create_port_on_network(network_N1)
        r1_id = self._create_router('r' + uuidutils.generate_uuid()[:4])
        self._attach_router_to_subnet(subnet_id, r1_id)

        # Create LB
        lb_data = self._create_load_balancer_and_validate(
            network_id, subnet_id, port_address, port_id, router_id=r1_id)
        # Create a pool
        self._create_pool_and_validate(lb_data, "p1")
        pool_id = lb_data['pools'][0].pool_id
        # Create Member-1 and associate it with lb_data
        self._create_member_and_validate(
            lb_data, pool_id, lb_data['vip_net_info'][1],
            lb_data['vip_net_info'][0], '10.0.0.10')
        # Create Member-2
        m_member = self._create_member_model(pool_id,
                                             lb_data['vip_net_info'][1],
                                             '10.0.0.12')
        # Update ovn's Logical switch reference
        self._update_ls_refs(lb_data, lb_data['vip_net_info'][0])
        lb_data['pools'][0].members.append(m_member)
        # Add a new member to the LB
        members = [m_member] + [lb_data['pools'][0].members[0]]
        self._update_members_in_batch_and_validate(lb_data, pool_id, members)
        # Deleting one member, while keeping the other member available
        self._update_members_in_batch_and_validate(lb_data, pool_id,
                                                   [m_member])
        # Create Member-3 with monitor options
        m_member = self._create_member_model(pool_id,
                                             lb_data['vip_net_info'][1],
                                             '10.0.0.12')
        m_member.monitor_port = 8080
        members = [m_member]
        self.assertRaises(o_exceptions.UnsupportedOptionError,
                          self.ovn_driver.member_batch_update,
                          pool_id,
                          members)
        self._delete_load_balancer_and_validate(lb_data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import exceptions as ks_exceptions
from keystoneauth1 import loading as ks_loading

from octavia_lib.api.drivers import exceptions as driver_exceptions
import openstack
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from ovn_octavia_provider.common import constants
from ovn_octavia_provider.i18n import _

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class KeystoneSession():

    def __init__(self, section=constants.SERVICE_AUTH):
        self._session = None
        self._auth = None

        self.section = section

    @property
    def session(self):
        """Initialize a Keystone session.

        :return: a Keystone Session object
        """
        if not self._session:
            self._session = ks_loading.load_session_from_conf_options(
                cfg.CONF, self.section, auth=self.auth)
        return self._session

    @property
    def auth(self):
        if not self._auth:
            try:
                self._auth = ks_loading.load_auth_from_conf_options(
                    cfg.CONF, self.section)
            except ks_exceptions.auth_plugins.MissingRequiredOptions as e:
                if self.section == constants.SERVICE_AUTH:
                    raise e
                # NOTE(gthiemonge): MissingRequiredOptions is raised: there is
                # one or more missing auth options in the config file. It may
                # be due to the migration from python-neutronclient to
                # openstacksdk.
                # With neutronclient, most of the auth settings were in
                # [service_auth] with a few overrides in [neutron],
                # but with openstacksdk, we have all the auth settings in the
                # [neutron] section. In order to support smooth upgrades, in
                # case those options are missing, we override the undefined
                # options with the existing settings from [service_auth].

                # This code should be removed when all the deployment tools set
                # the correct options in [neutron]

                # The config options are lazily registered/loaded by keystone,
                # it means that we cannot get/set them before invoking
                # 'load_auth_from_conf_options' on 'service_auth'.
                ks_loading.load_auth_from_conf_options(
                    cfg.CONF, constants.SERVICE_AUTH)

                config = getattr(cfg.CONF, self.section)
                for opt in config:
                    # For each option in the [neutron] section, get its setting
                    # location, if the location is 'opt_default' or
                    # 'set_default', it means that the option is not configured
                    # in the config file, it should be replaced with the one
                    # from [service_auth]
                    loc = cfg.CONF.get_location(opt, self.section)
                    if not loc or loc.location in (cfg.Locations.opt_default,
                                                   cfg.Locations.set_default):
                        if hasattr(cfg.CONF.service_auth, opt):
                            cur_value = getattr(config, opt)
                            value = getattr(cfg.CONF.service_auth, opt)
                            if value != cur_value:
                                log_value = (value if opt != "password"
                                             else "<hidden>")
                                LOG.debug("Overriding [%s].%s with '%s'",
                                          self.section, opt, log_value)
                                cfg.CONF.set_override(opt, value, self.section)

                # Now we can call load_auth_from_conf_options for this specific
                # service with the newly defined options.
                self._auth = ks_loading.load_auth_from_conf_options(
                    cfg.CONF, self.section)

        return self._auth


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args,
                                                                 **kwargs)
        return cls._instances[cls]


class NeutronAuth(metaclass=Singleton):
    def __init__(self):
        """Create neutron client object."""
        try:
            ksession = KeystoneSession('neutron')

            kwargs = {}
            if CONF.neutron.endpoint_override:
                kwargs['network_endpoint_override'] = (
                    CONF.neutron.endpoint_override)

            self.network_proxy = openstack.connection.Connection(
                session=ksession.session, **kwargs).network
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Error creating Neutron client.")


def get_neutron_client():
    try:
        return NeutronAuth().network_proxy
    except Exception as e:
        msg = _('Cannot initialize OpenStackSDK. Exception: %s. '
                'Please verify Neutron service configuration '
                'in Octavia API configuration.') % e
        raise driver_exceptions.DriverError(
            operator_fault_string=msg)


class OctaviaAuth(metaclass=Singleton):
    def __init__(self):
        """Create Octavia client object."""
        try:
            ksession = KeystoneSession()
            kwargs = {'region_name': CONF.service_auth.region_name}
            # TODO(ricolin) `interface` option don't take list as option yet.
            # We can move away from this when openstacksdk no longer depends
            # on `interface`.
            try:
                interface = CONF.service_auth.valid_interfaces[0]
            except (TypeError, LookupError):
                interface = CONF.service_auth.valid_interfaces
            if interface:
                kwargs['interface'] = interface

            self.loadbalancer_proxy = openstack.connection.Connection(
                session=ksession.session, **kwargs).load_balancer
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Error creating Octavia client.")


def get_octavia_client():
    try:
        return OctaviaAuth().loadbalancer_proxy
    except Exception as e:
        msg = _('Cannot initialize OpenStackSDK. Exception: %s. '
                'Please verify service_auth configuration '
                'in Octavia API configuration.') % e
        raise driver_exceptions.DriverError(
            operator_fault_string=msg)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest import mock

from ovn_octavia_provider import agent as ovn_agent
from ovn_octavia_provider.tests.unit import base as ovn_base


class TestOvnProviderAgent(ovn_base.TestOvnOctaviaBase):

    def test_exit(self):
        mock_exit_event = mock.MagicMock()
        mock_exit_event.is_set.side_effect = [False, False, False, False, True]
        ovn_agent.OvnProviderAgent(mock_exit_event)
        self.assertEqual(1, mock_exit_event.wait.call_count)
        self.assertEqual(3, self.mock_ovn_nb_idl.call_count)
        self.assertEqual(1, self.mock_ovn_sb_idl.call_count)
